def xzbackdoor(n, e=0):  # noqa: ARG001
    byte_len = (n.bit_length() + 7) // 8
    # avoid error with malformed N=0 RSA key
    if byte_len == 0:
        return False

    nb = n.to_bytes(byte_len, byteorder="big")
    # if upper bit is set, ASN1 encodes with leading zero byte
    if nb[0] & 0x80:
        nb = b"\x00" + nb

    a = int.from_bytes(nb[0:4], byteorder="little", signed=False)
    b = int.from_bytes(nb[4:8], byteorder="little", signed=False)
    c = int.from_bytes(nb[8:16], byteorder="little", signed=False)
    res = 0xFFFFFFFFFFFFFFFF & (a * b + c)
    if res <= 3:
        return {"detected": True}
    return False



# SPDX-License-Identifier: MIT
# Copyright (c) 2017, CRoCS, EnigmaBridge Ltd.
# Copyright (c) Hanno Böck
#
# Original: https://github.com/crocs-muni/roca
# Reduced to core functionality by Hanno Böck

import gmpy2


def roca(element, e=0):  # noqa: ARG001
    generator = 65537
    generator_order = 2454106387091158800
    pp = [16, 81, 25, 7, 11, 13, 17, 23, 29, 37, 41, 53, 83]
    modulus = 0x924CBA6AE99DFA084537FACC54948DF0C23DA044D8CABE0EDD75BC6
    if element <= 2:
        return False
    if gmpy2.powmod(element, generator_order, modulus) != 1:
        return False

    for prime_to_power in pp:
        order_div_prime_power = generator_order // prime_to_power
        g_dash = gmpy2.powmod(generator, order_div_prime_power, modulus)
        h_dash = gmpy2.powmod(element, order_div_prime_power, modulus)
        found = False
        for i in range(prime_to_power):
            if gmpy2.powmod(g_dash, i, modulus) == h_dash:
                found = True
                break
        if not found:
            return False
    return {"detected": True}



import hashlib
import warnings

import cryptography
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import dh, dsa, ec, ed448, ed25519, rsa, x448, x25519

from .allkeys import blocklist
from .rsakeys import (fermat, pattern, roca, rsainvalid, rsarecover, rsawarnings, sharedprimes,
                      smalld, smallfactors, xzbackdoor)

# List of available checks
defaultchecks = {
    "fermat": {
        "type": "rsa",
        "function": fermat,
        "desc": "Fermat factorization / close prime vulnerability",
    },
    "pattern": {
        "type": "rsa",
        "function": pattern,
        "desc": "Implausible repetition pattern in modulus",
    },
    "roca": {
        "type": "rsa",
        "function": roca,
        "desc": "Return of the Coopersmith Attack (ROCA) vulnerability",
    },
    "rsainvalid": {
        "type": "rsa",
        "function": rsainvalid,
        "desc": "RSA keys with invalid values",
    },
    "sharedprimes": {
        "type": "rsa",
        "function": sharedprimes,
        "desc": "Shared prime factors (batchgcd)",
    },
    "smalld": {
        "type": "rsa",
        "function": smalld,
        "desc": "Small private d (Wiener's attack)",
    },
    "smallfactors": {
        "type": "rsa",
        "function": smallfactors,
        "desc": "Small prime factors (<=65537, usually corrupt)",
    },
    "xzbackdoor": {
        "type": "rsa",
        "function": xzbackdoor,
        "desc": "xz backdoor payload in RSA n",
    },
    "blocklist": {
        "type": "all",
        "function": blocklist,
        "desc": "Blocklists of compromised keys",
    },
}

warningchecks = {
    "rsawarnings": {
        "type": "rsa",
        "function": rsawarnings,
        "desc": "RSA key size and exponent warnings",
    },
}

allchecks = defaultchecks | warningchecks

# cryptography warns about SSH DSA keys being deprecated.
# For now, disable the warnings. Needs a better long-term solution.
warnings.filterwarnings(
    "ignore",
    category=cryptography.utils.CryptographyDeprecationWarning,
    module="badkeys.checks",
)


def _checkkey(key, checks, keyrecover=False):
    r = {}
    if isinstance(key, rsa.RSAPublicKey):
        r["type"] = "rsa"
        r["n"] = key.public_numbers().n
        r["e"] = key.public_numbers().e
        r["bits"] = r["n"].bit_length()
        r["results"] = checkrsa(r["n"], e=r["e"], checks=checks, keyrecover=keyrecover)
    elif isinstance(key, ec.EllipticCurvePublicKey):
        r["type"] = "ec"
        r["curve"] = key.curve.name
        # shorter names for standard curves
        if r["curve"].startswith("sec") and r["curve"].endswith("r1"):
            r["curve"] = r["curve"][3:-2]
        r["x"] = key.public_numbers().x
        r["y"] = key.public_numbers().y
        r["results"] = checkall(r["x"], checks=checks)
    elif isinstance(key, dsa.DSAPublicKey):
        r["type"] = "dsa"
        r["y"] = key.public_numbers().y
        r["bits"] = key.key_size
        r["results"] = checkall(r["y"], checks=checks)
    elif isinstance(
        key,
        (
            ed25519.Ed25519PublicKey,
            x25519.X25519PublicKey,
            x448.X448PublicKey,
            ed448.Ed448PublicKey,
        ),
    ):
        r["type"] = "ec"
        r["curve"] = str(type(key).__name__).lower()[:-9]
        # convert the raw key into an integer for the blocklist check
        pub_b = key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        )
        r["pub"] = int.from_bytes(pub_b, byteorder="big")
        r["results"] = checkall(r["pub"], checks=checks)
    elif isinstance(key, dh.DHPublicKey):
        r["type"] = "dh"
        try:
            r["y"] = key.public_numbers().y
        except ValueError:
            # happens with e.g. very small (<512) DH keys
            return {"type": "unparseable", "results": {}}
        r["results"] = checkall(r["y"], checks=checks)
    else:
        r["type"] = "unsupported"
        r["results"] = {}
    spki = key.public_bytes(
        serialization.Encoding.DER,
        serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    r["spkisha256"] = hashlib.sha256(spki).hexdigest()
    return r


def checkrsa(n, e=65537, checks=defaultchecks.keys(), keyrecover=False):
    results = {}
    for check in checks:
        callcheck = allchecks[check]["function"]
        if allchecks[check]["type"] == "rsa":
            r = callcheck(n, e=e)
        elif allchecks[check]["type"] == "all":
            r = callcheck(n)
        else:
            continue
        if r is not False:
            pemkey = None
            if keyrecover and "p" in r and "q" in r:
                pemkey = rsarecover(p=r["p"], q=r["q"], n=n, e=e)
            elif keyrecover and "d" in r:
                pemkey = rsarecover(d=r["d"], n=n, e=e)
            if pemkey:
                r["privatekey"] = pemkey
            results[check] = r
    return results


def checkall(x, checks=defaultchecks.keys()):
    results = {}
    for check in checks:
        if allchecks[check]["type"] != "all":
            continue
        callcheck = allchecks[check]["function"]
        r = callcheck(x)
        if r is not False:
            results[check] = r
    return results


def _reterr(rtype, ex):
    return {"type": rtype, "results": {}, "exception": ex.__class__.__name__, "errmsg": str(ex)}


def checkpubkey(rawkey, checks=defaultchecks.keys(), keyrecover=False):
    try:
        key = serialization.load_pem_public_key(rawkey.encode())
    except ValueError as e:
        # ValueError: defect inputs (partial keys, ASN.1 errors), ECDSA with explicit curves,
        #             unknown key types (e.g., ML-DSA)
        return _reterr("unparseable", e)
    except cryptography.exceptions.UnsupportedAlgorithm as e:
        # UnsupportedAlgorithm: Unsupported elliptic curves
        return _reterr("unsupported", e)
    return _checkkey(key, checks, keyrecover=keyrecover)


def checkprivkey(rawkey, checks=defaultchecks.keys(), keyrecover=False):
    try:
        priv = serialization.load_pem_private_key(rawkey.encode(), password=None)
    except ValueError as e:
        # ValueError: defect inputs, unknown key types, explicit curves
        return _reterr("unparseable", e)
    except (cryptography.exceptions.UnsupportedAlgorithm, TypeError) as e:
        # UnsupportedAlgorithm: Unsupported elliptic curves
        # TypeError: passwords-protected keys
        return _reterr("unsupported", e)
    return _checkkey(priv.public_key(), checks, keyrecover=keyrecover)


def checkcrt(rawcert, checks=defaultchecks.keys(), keyrecover=False):
    try:
        crt = x509.load_pem_x509_certificate(rawcert.encode())
    except (ValueError, cryptography.x509.base.InvalidVersion) as e:
        return _reterr("unparseable", e)
    try:
        pubkey = crt.public_key()
    except (cryptography.exceptions.UnsupportedAlgorithm, ValueError, NotImplementedError) as e:
        # ValueError: unknown key types, explicit curves
        # UnsupportedAlgorithm: unsupported curves
        # NotImplementedError: ? (possibly certificate extension issues)
        return _reterr("unsupported", e)
    return _checkkey(pubkey, checks, keyrecover=keyrecover)


def checkcsr(rawcsr, checks=defaultchecks.keys(), keyrecover=False):
    try:
        csr = x509.load_pem_x509_csr(rawcsr.encode())
    except (ValueError, cryptography.x509.base.InvalidVersion) as e:
        return _reterr("unparseable", e)
    try:
        pubkey = csr.public_key()
    except (cryptography.exceptions.UnsupportedAlgorithm, ValueError) as e:
        # ValueError: unknown key types, explicit curves
        # UnsupportedAlgorithm: unsupported curves
        return _reterr("unsupported", e)
    return _checkkey(pubkey, checks, keyrecover=keyrecover)


def checksshprivkey(sshkey, checks=defaultchecks.keys(), keyrecover=False):
    try:
        pkey = serialization.load_ssh_private_key(sshkey.encode(), password=None)
    except ValueError as e:
        # ValueError: defect keys
        return _reterr("unparseable", e)
    except (cryptography.exceptions.UnsupportedAlgorithm, TypeError) as e:
        # UnsupportedAlgorithm: unsupported key types, e.g., sk-*
        # TypeError: password-protected keys
        return _reterr("unsupported", e)
    return _checkkey(pkey.public_key(), checks, keyrecover=keyrecover)


def checksshpubkey(sshkey, checks=defaultchecks.keys(), keyrecover=False):
    try:
        pkey = serialization.load_ssh_public_key(sshkey.encode())
    except ValueError as e:
        # ValueError: defect keys, non-standard DSA keys (!=1024 bit)
        return _reterr("unparseable", e)
    except (cryptography.exceptions.UnsupportedAlgorithm, NotImplementedError) as e:
        # UnsupportedAlgorithm: unsupported key types
        # NotImplementedError: unsupported EC point format
        return _reterr("unsupported", e)
    return _checkkey(pkey, checks, keyrecover=keyrecover)


def detectandcheck(inkey, checks=defaultchecks.keys(), keyrecover=False):
    if "-----BEGIN CERTIFICATE-----" in inkey:
        return checkcrt(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN CERTIFICATE REQUEST-----" in inkey:
        return checkcsr(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN PUBLIC KEY-----" in inkey:
        return checkpubkey(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN RSA PUBLIC KEY-----" in inkey:
        return checkpubkey(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN RSA PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN DSA PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN EC PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks, keyrecover=keyrecover)
    if "-----BEGIN OPENSSH PRIVATE KEY-----" in inkey:
        return checksshprivkey(inkey, checks, keyrecover=keyrecover)
    if inkey.startswith(("ssh-", "ecdsa-")):
        return checksshpubkey(inkey, checks, keyrecover=keyrecover)
    return {"type": "notfound", "results": {}}



import base64
import binascii

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa


def parsedkim(line):

    # remove escaped quote characters, they can break our parser
    line = line.replace('\\"', "")

    qsplit = line.split('"')
    # Concat quoted strings if we have double quotes.
    # Exception: if we have a '=' before the first quote it is not
    # a quoted record, but quotes within a DKIM value (see example
    # in tests/data/dkim/dkim-comment-quotes.txt).
    if len(qsplit) >= 2 and "=" not in qsplit[0]:
        add = 0
        dk = ""
        for x in line.split('"'):
            add ^= 1
            if add == 0:
                dk += x
    else:
        dk = line

    dkim = {}
    for x in dk.split(";"):
        s = x.split("=", 1)
        if len(s) != 2:
            continue
        key = s[0].strip()
        value = s[1].strip()
        dkim[key] = value
    if "p" not in dkim:
        return False
    if dkim["p"] == "":
        return False
    if "k" not in dkim:
        dkim["k"] = "rsa"

    if dkim["k"] == "rsa":
        try:
            rawkey = base64.b64decode(dkim["p"].encode())
        except (binascii.Error, UnicodeEncodeError):
            return "base64error"
        try:
            pubkey = serialization.load_der_public_key(rawkey)
        except ValueError:
            try:
                serialization.load_der_private_key(rawkey, password=None)
                return "privleak"
            except ValueError:
                pass
            return "keyparseerror"
        if not isinstance(pubkey, rsa.RSAPublicKey):
            return "expectedrsa"
        return pubkey
    if dkim["k"] == "ed25519":
        try:
            rawed = base64.b64decode(dkim["p"].encode())
        except (binascii.Error, UnicodeEncodeError):
            return "base64error"
        if len(rawed) != 32:
            return "wrongkeylength"
        return ed25519.Ed25519PublicKey.from_public_bytes(rawed)
    return "unknowntype"



import hashlib
import json
import lzma
import os
import pathlib
import string
import urllib.request

from .utils import _cachedir, _errexit, _warnmsg


def _dlxz(url, filename, exphash, quiet):
    if not quiet:
        print(f"Downloading {filename}...")
    dldata = urllib.request.urlopen(url).read()
    dlunpacked = lzma.decompress(dldata)
    dlhash = hashlib.sha256(dlunpacked).hexdigest()
    if dlhash != exphash:
        _errexit(f"SHA256 hash of downloaded {filename} does not match")
    tmpfile = os.path.join(_cachedir(), f"_{filename}.tmp")
    pathlib.Path(tmpfile).write_bytes(dlunpacked)
    os.replace(tmpfile, os.path.join(_cachedir(), filename))


def update_bl(lookup=False, quiet=False):
    UPDATEURL = "https://update.badkeys.info/"
    BKFORMAT = 0

    if not os.path.exists(_cachedir()):
        pathlib.Path(_cachedir()).mkdir(parents=True)

    jurl = f"{UPDATEURL}/v{BKFORMAT}/badkeysdata.json"
    bkdata = urllib.request.urlopen(jurl).read()
    bkdata_old = ""
    bkjsonfile = os.path.join(_cachedir(), "badkeysdata.json")
    if os.path.exists(bkjsonfile):
        bkdata_old = pathlib.Path(bkjsonfile).read_bytes()

    data = json.loads(bkdata)
    if data["bkformat"] != BKFORMAT:
        _errexit("Wrong format")

    if "deprecated" in data:
        _warnmsg("Your badkeys version uses a deprecated blocklist format, please update!")
        # "deprecated" field can be 'true' or a string with additional info
        if isinstance(data["deprecated"], str):
            # make sure message is pure ASCII without ANSI control chars
            depmsg = "".join(filter(lambda x: x in string.printable, data["deprecated"]))
            _warnmsg(depmsg)

    if "blocklist_url" not in data:
        _warnmsg("Blocklist no longer available in requested format")
        return

    if bkdata == bkdata_old:
        if not quiet:
            print("No new data")
    else:
        if not quiet:
            print("Writing new badkeysdata.json...")
        pathlib.Path(bkjsonfile).write_bytes(bkdata)

    # starting with blocklist.dat
    oldbl_sha256 = ""
    if os.path.exists(os.path.join(_cachedir(), "blocklist.dat")):
        with open(os.path.join(_cachedir(), "blocklist.dat"), "rb") as f:
            oldbl_sha256 = hashlib.sha256(f.read()).hexdigest()

    if oldbl_sha256 != data["blocklist_sha256"]:
        _dlxz(data["blocklist_url"], "blocklist.dat", data["blocklist_sha256"], quiet)

    # starting with lookup.txt
    oldlu_sha256 = ""
    if os.path.exists(os.path.join(_cachedir(), "lookup.txt")):
        with open(os.path.join(_cachedir(), "lookup.txt"), "rb") as f:
            oldlu_sha256 = hashlib.sha256(f.read()).hexdigest()
        if not lookup and (oldlu_sha256 != data["lookup_sha256"]):
            _warnmsg("Old lookup.txt file found.")
            _warnmsg("You may want to run --update-bl-and-urls")

    if lookup and (oldlu_sha256 != data["lookup_sha256"]):
        _dlxz(data["lookup_url"], "lookup.txt", data["lookup_sha256"], quiet)



__all__ = [
    "fermat",
    "pattern",
    "roca",
    "rsainvalid",
    "rsarecover",
    "rsawarnings",
    "sharedprimes",
    "smalld",
    "smallfactors",
    "xzbackdoor",
]
from .fermat import fermat
from .pattern import pattern
from .roca import roca
from .rsainvalid import rsainvalid
from .rsarecover import rsarecover
from .rsawarnings import rsawarnings
from .sharedprimes import sharedprimes
from .smalld import smalld
from .smallfactors import smallfactors
from .xzbackdoor import xzbackdoor



# SPDX-License-Identifier: MIT
# Copyright (c) Hanno Böck

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa


def rsarecover(p=None, q=None, d=None, n=None, e=65537):
    try:
        if p and q:
            d = rsa.rsa_recover_private_exponent(e, p, q)
        elif d:
            p, q = rsa.rsa_recover_prime_factors(n, e, d)
        else:
            return None
        dmp1 = rsa.rsa_crt_dmp1(d, p)
        dmq1 = rsa.rsa_crt_dmq1(d, q)
        iqmp = rsa.rsa_crt_iqmp(p, q)
        pubnum = rsa.RSAPublicNumbers(e, n)
        privnum = rsa.RSAPrivateNumbers(p, q, d, dmp1, dmq1, iqmp, pubnum)
        privkey = privnum.private_key()
        pemprivate = privkey.private_bytes(serialization.Encoding.PEM,
                                           serialization.PrivateFormat.TraditionalOpenSSL,
                                           serialization.NoEncryption())
    except ValueError:
        return None
    return pemprivate.decode("ascii")



import functools
import os
import sys

_retval = 0


def _esc(inp):
    return repr(inp)[1:-1]


@functools.cache
def _setret(rv):
    global _retval
    _retval |= rv


def _getret():
    return _retval


def _warnmsg(msg):
    _setret(2)
    print(f"WARNING: {msg}", file=sys.stderr)


def _errexit(msg):
    print(f"ERROR: {msg}", file=sys.stderr)
    sys.exit(_retval | 1)


@functools.cache
def _cachedir():
    cachedir = os.getenv("XDG_CACHE_HOME")
    if cachedir:
        return os.path.join(cachedir, "badkeys", "")
    return os.path.expanduser("~/.cache/badkeys/")



def rsawarnings(n, e=65537):
    bits = n.bit_length()

    # In theory, multiple warnings could apply, but badkeys
    # currently cannot represent this.
    # We start with the most severe, and go down from there.
    if bits < 768:
        return {"detected": True, "subtest": "extremely_small"}
    if bits < 2048:
        return {"detected": True, "subtest": "too_small"}
    if (bits % 8) != 0:
        return {"detected": True, "subtest": "not_multiple_of_8"}
    if bits not in [2048, 3072, 4096]:
        return {"detected": True, "subtest": "unusual_keysize"}

    if e == 3:
        return {"detected": True, "subtest": "exponent_3"}
    if e != 65537:
        return {"detected": True, "subtest": "exponent_not_65537"}

    return False



import hashlib
import json
import mmap
import os

from badkeys.utils import _cachedir, _errexit

_blmeta = False
_bldata = False
_blextra = []


def _loadblmeta():
    global _blmeta
    mlist = {}
    try:
        with open(os.path.join(_cachedir(), "badkeysdata.json")) as f:
            jdata = json.loads(f.read())
    except FileNotFoundError:
        _errexit("blocklist metadata not found, you need to run --update-bl")
    for bl in jdata["blocklists"]:
        blid = int(bl["id"])
        mlist[blid] = bl
    _blmeta = mlist


def _checkbl(bldata, s256trunc):
    fbegin = 0
    fend = (len(bldata) // 16) - 1
    while fbegin <= fend:
        fmiddle = (fbegin + fend) // 2
        val = bldata[fmiddle * 16 : fmiddle * 16 + 15]
        if s256trunc == val:
            bl_id = int(bldata[fmiddle * 16 + 15])
            if bl_id in _blmeta:
                subtest = _blmeta[bl_id]["name"]
            else:
                subtest = f"id{bl_id}"
            lhash = s256trunc[0:8].hex()
            return {
                "detected": True,
                "subtest": subtest,
                "blid": bl_id,
                "lookup": lhash,
                "debug": f"Truncated Hash: {s256trunc.hex()}",
            }
        if s256trunc > val:
            fbegin = fmiddle + 1
        else:
            fend = fmiddle - 1

    return False


def blocklist(inval):
    global _bldata

    if not _blmeta:
        _loadblmeta()

    if not _bldata:
        try:
            with open(os.path.join(_cachedir(), "blocklist.dat"), "rb") as f:
                _bldata = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        except FileNotFoundError:
            _errexit("blocklist.dat not found, you need to run --update-bl")

    inval_b = inval.to_bytes((inval.bit_length() + 7) // 8, byteorder="big")

    s256trunc = hashlib.sha256(inval_b).digest()[:15]

    ret = _checkbl(_bldata, s256trunc)
    if ret:
        return ret

    for bl in _blextra:
        ret = _checkbl(bl, s256trunc)
        if ret:
            return ret

    return False


def loadextrabl(fpath):
    with open(fpath, "rb") as f:
        _blextra.append(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ))


def urllookup(blid, lhash):
    try:
        from binary_file_search.BinaryFileSearch import BinaryFileSearch
    except ModuleNotFoundError:
        _errexit("URL lookup failed, needs binary_file_search module")

    if not _blmeta:
        _loadblmeta()

    lfile = os.path.join(_cachedir(), "lookup.txt")

    try:
        with BinaryFileSearch(lfile, sep=";", string_mode=True) as bfs:
            x = bfs.search(lhash)
    except FileNotFoundError:
        _errexit("lookup.txt not found, run --update-bl-and-urls")
    except KeyError:
        return None, None
    d = _blmeta[blid]
    showurl = f"https://github.com/{d['repo']}/blob/{d['path']}/{x[0][1]}"
    rawurl = f"https://raw.githubusercontent.com/{d['repo']}/{d['path']}/{x[0][1]}"
    return showurl, rawurl



import base64
import binascii

from .checks import checkall, checkrsa


def checkdnskey(rec, checks):
    o = rec.split(maxsplit=3)
    if len(o) != 4:
        return {"type": "unparseable", "results": {}}
    try:
        keytype = int(o[2])
    except ValueError:
        return {"type": "unparseable", "results": {}}
    try:
        key = base64.b64decode(o[3].replace(" ", "").encode())
    except (binascii.Error, UnicodeEncodeError):
        return {"type": "unparseable", "results": {}}
    r = {}
    if keytype in {1, 5, 7, 8, 10}:  # RSA
        # RSA key format description in RFC 3110 Section 2
        if len(key) < 3:
            return {"type": "unparseable", "results": {}}
        if key[0] == 0:
            elen = int.from_bytes(key[1:3], byteorder="big")
            eoffset = 3
        else:
            elen = key[0]
            eoffset = 1

        if len(key) < elen + eoffset:
            return {"type": "unparseable", "results": {}}

        r["type"] = "rsa"
        r["e"] = int.from_bytes(key[eoffset:eoffset + elen], byteorder="big")
        r["n"] = int.from_bytes(key[eoffset + elen:], byteorder="big")
        r["bits"] = r["n"].bit_length()
        r["results"] = checkrsa(r["n"], r["e"], checks=checks)
        return r
    if keytype in {3, 6}:  # DSA
        # DSA key format description in RFC 2536 Section 2
        if len(key) < 213:
            return {"type": "unparseable", "results": {}}
        dsa_t = int(key[0])
        if dsa_t > 8 or len(key) != 213 + dsa_t * 24:
            return {"type": "unparseable", "results": {}}
        r["type"] = "dsa"
        y_off = 149 + dsa_t * 16
        r["y"] = int.from_bytes(key[y_off:], "big")
        r["bits"] = 512 + dsa_t * 64
        r["results"] = checkall(r["y"], checks=checks)
        return r
    if keytype == 13:  # ECDSAP256SHA256
        if len(key) != 64:
            return {"type": "unparseable", "results": {}}
        r["curve"] = "p256"
        r["x"] = int.from_bytes(key[0:32], byteorder="big")
    if keytype == 14:  # ECDSAP384SHA384
        if len(key) != 96:
            return {"type": "unparseable", "results": {}}
        r["curve"] = "p384"
        r["x"] = int.from_bytes(key[0:48], byteorder="big")
    if keytype == 15:  # ED25519
        if len(key) != 32:
            return {"type": "unparseable", "results": {}}
        r["curve"] = "ed25519"
        r["x"] = int.from_bytes(key, byteorder="big")
    if keytype == 16:  # ED448
        if len(key) != 57:
            return {"type": "unparseable", "results": {}}
        r["curve"] = "ed448"
        r["x"] = int.from_bytes(key, byteorder="big")
    if "x" in r:
        r["type"] = "ec"
        r["results"] = checkall(r["x"], checks=checks)
        return r
    return {"type": "unsupported", "results": {}}



import ssl

from .checks import checkcrt, defaultchecks


def scantls(host, port=443, checks=defaultchecks.keys()):
    try:
        # this currently does not support multiple
        # certificates on one host
        cert = ssl.get_server_certificate((host, port))
    except (ConnectionRefusedError, ssl.SSLError, OSError):
        return []
    return [checkcrt(cert, checks=checks)]



__version__ = "0.0.17"
__all__ = [
    "allchecks",
    "checkcrt",
    "checkcsr",
    "checkdnskey",
    "checkjwk",
    "checkprivkey",
    "checkpubkey",
    "checkrsa",
    "checksshpubkey",
    "defaultchecks",
    "detectandcheck",
    "scanssh",
    "warningchecks",
]
from .checks import (allchecks, checkcrt, checkcsr, checkprivkey, checkpubkey, checkrsa,
                     checksshpubkey, defaultchecks, detectandcheck, warningchecks)
from .dnssec import checkdnskey
from .jwk import checkjwk
from .scanssh import scanssh



import base64
import binascii

from .checks import checkall, checkrsa

# See https://www.iana.org/assignments/jose/jose.xhtml
VALIDCURVES = ["P-256", "P-384", "P-521", "Ed25519", "Ed448", "X25519", "X448", "secp256k1"]


def _ub64toint(b64):
    fb64 = b64.replace(" ", "").encode()
    pad = b"=" * ((-len(fb64)) % 4)
    raw = base64.urlsafe_b64decode(fb64 + pad)
    return int.from_bytes(raw, byteorder="big")


def checkjwk(key, checks):
    r = {}
    if not isinstance(key, dict) or "kty" not in key:
        return {"type": "unparseable", "results": {}}
    try:
        if key["kty"] == "RSA":
            if "n" not in key or "e" not in key or key["n"] == "" or key["e"] == "" \
               or not isinstance(key["n"], str) or not isinstance(key["e"], str):
                return {"type": "unparseable", "results": {}}
            r["type"] = "rsa"
            r["n"] = _ub64toint(key["n"])
            r["e"] = _ub64toint(key["e"])
            r["bits"] = r["n"].bit_length()
            r["results"] = checkrsa(r["n"], e=r["e"], checks=checks)
        elif key["kty"] == "EC":
            if "x" not in key or "y" not in key or key["x"] == "" or key["y"] == "" \
               or not isinstance(key["x"], str) or not isinstance(key["y"], str) \
               or "crv" not in key or key["crv"] not in VALIDCURVES:
                return {"type": "unparseable", "results": {}}
            r["type"] = "ec"
            r["curve"] = key["crv"].lower().replace("-", "")
            r["x"] = _ub64toint(key["x"])
            r["y"] = _ub64toint(key["x"])
            r["results"] = checkall(r["x"], checks=checks)
        elif key["kty"] == "OKP":
            if "x" not in key or key["x"] == "" \
               or not isinstance(key["x"], str) \
               or "crv" not in key or key["crv"] not in VALIDCURVES:
                return {"type": "unparseable", "results": {}}
            r["type"] = "ec"
            r["curve"] = key["crv"].lower()
            r["x"] = _ub64toint(key["x"])
            # no y coordinate for ed25519/ed448
            r["results"] = checkall(r["x"], checks=checks)
        else:
            return {"type": "unsupported", "results": {}}
    except (binascii.Error, UnicodeEncodeError):
        return {"type": "unparseable", "results": {}}
    return r



import re

# Find suspicious patterns of 16 repeating bytes
_bprex = re.compile(rb"(.)\1{15}")


def pattern(n, e=0):  # noqa: ARG001
    nbin = n.to_bytes((n.bit_length() + 7) // 8, byteorder="big")
    if _bprex.search(nbin):
        return {"detected": True}
    return False



# SPDX-License-Identifier: MIT
# (c) Nao Yonashiro
# (c) Hanno Böck
#
# Part of badkeys: https://badkeys.info/
#
# Based on:
# https://github.com/orisano/owiener/

from collections.abc import Iterable, Iterator

import gmpy2


def rational_to_contfrac(x: int, y: int) -> Iterator[int]:
    """
    ref: https://en.wikipedia.org/wiki/Euclidean_algorithm#Continued_fractions

    >>> list(rational_to_contfrac(4, 11))
    [0, 2, 1, 3]
    """
    while y:
        a = x // y
        yield a
        x, y = y, x - a * y


def contfrac_to_rational_iter(contfrac: Iterable[int]) -> Iterator[tuple[int, int]]:
    n0, d0 = 0, 1
    n1, d1 = 1, 0
    for q in contfrac:
        n = q * n1 + n0
        d = q * d1 + d0
        yield n, d
        n0, d0 = n1, d1
        n1, d1 = n, d


def convergents_from_contfrac(contfrac: Iterable[int]) -> Iterator[tuple[int, int]]:
    n_, d_ = 1, 0
    for i, (n, d) in enumerate(contfrac_to_rational_iter(contfrac)):
        if i % 2 == 0:
            yield n + n_, d + d_
        else:
            yield n, d
        n_, d_ = n, d


def smalld(n, e):
    # it makes no sense to test with small e
    if e.bit_length() <= 32:
        return False
    f_ = rational_to_contfrac(e, n)
    for k, dg in convergents_from_contfrac(f_):
        edg = e * dg
        phi = edg // k

        x = n - phi + 1
        if x % 2 == 0 and gmpy2.is_square((x // 2) ** 2 - n):
            g = edg - phi * k
            return {"d": dg // g}
    return False



import functools

import gmpy2

MAX_PRIME = 65537


@functools.cache
def _gensmallprimes():
    # Generate product of all primes <= MAX_PRIME.
    # We calculate this once per program run, we could precalculate the
    # constant, but it's fast enough to calculate on the fly.
    sp = prime = 2
    while prime < MAX_PRIME:
        prime = gmpy2.next_prime(prime)
        sp *= prime
    return sp


def smallfactors(n, e=0):  # noqa: ARG001
    # Don't try to factor nonsensical keys
    if n <= 5:
        return False

    factor = gmpy2.gcd(_gensmallprimes(), n)
    if factor == 1:
        return False
    p = factor
    q = n // factor
    if not gmpy2.is_prime(p) or not gmpy2.is_prime(q):
        sub = "corrupt"
    else:
        sub = "valid"

    # convert gmpy2 mpz to python integers
    return {"p": int(p), "q": int(q), "subtest": sub}



import argparse
import json
import re
import signal
import sys

from . import __version__
from .allkeys import loadextrabl, urllookup
from .checks import (_checkkey, allchecks, checkcrt, checkrsa, checksshpubkey, defaultchecks,
                     detectandcheck)
from .dkim import parsedkim
from .dnssec import checkdnskey
from .jwk import checkjwk
from .scanssh import scanssh
from .scantls import scantls
from .update import update_bl
from .utils import _errexit, _esc, _getret, _setret, _warnmsg

MAXINPUTSIZE = 2048000

count = 0

PRECRT = "-----BEGIN CERTIFICATE-----\n"
POSTCRT = "\n-----END CERTIFICATE-----\n"

_parseerrmsg = {
    "privleak": "Private instead of public key",
    "keyparseerror": "Error decoding DER/ASN.1 key structure",
    "expectedrsa": "Wrong key type, expected RSA key",
    "wrongkeylength": "Wrong key length",
    "unknowntype": "Unknown key type",
}


def _sighandler(_signum, _handler):
    print(f"{count} keys processed", file=sys.stderr)


def _printresults(key, where, args):
    if args.json:
        jout = key
        jout["where"] = where
        for val in ["n", "e", "x", "y"]:
            if val in jout:
                jout[val] = f"{jout[val]:x}"
        for result in jout["results"]:
            for val in ["p", "q", "e"]:
                if val in jout["results"][result]:
                    jout["results"][result][val] = f'{jout["results"][result][val]:x}'
        if args.all or jout["results"] or \
           key["type"] in ["unsupported", "unparseable", "notfound"]:
            print(json.dumps(jout))
        return
    kn = key["type"]
    if "bits" in key:
        kn += f"[{key['bits']}]"
    if "curve" in key:
        kn += f"[{key['curve']}]"
    if key["type"] == "unsupported":
        _warnmsg(f"Unsupported key type, {_esc(where)}")
    elif key["type"] == "unparseable":
        reason = ""
        if reason in key and key["reason"] in _parseerrmsg:
            reason = f" ({_parseerrmsg[key['reason']]})"
        elif "reason" in key:
            reason = f" ({key['reason']})"
        _warnmsg(f"Unparseable input{reason}, {_esc(where)}")
    elif key["type"] == "notfound":
        _warnmsg(f"No key found, {_esc(where)}")
    elif args.verbose or args.all:
        if key["results"] == {}:
            print(f"{kn} key ok, {_esc(where)}")
    for check, result in key["results"].items():
        sub = ""
        if "subtest" in result:
            sub = f"/{result['subtest']}"
        print(f"{check}{sub} vulnerability, {kn}, {_esc(where)}")
        _setret(4)
        if args.url and "lookup" in result:
            url, _ = urllookup(result["blid"], result["lookup"])
            if url:
                print(url)
            else:
                _warnmsg("URL lookup failed, not found")
        if args.verbose and "debug" in result:
            print(result["debug"])
        if args.verbose and "p" in result:
            print(f"RSA p {result['p']:02x}")
        if args.verbose and "q" in result:
            print(f"RSA q {result['q']:02x}")


def runcli():
    global count
    try:
        signal.signal(signal.SIGHUP, _sighandler)
    except AttributeError:  # for OSes without SIGHUP (e.g. Windows)
        pass

    ap = argparse.ArgumentParser()
    ap.add_argument("infiles", nargs="*", help="Input file (certificate, csr or public key)")
    ap.add_argument("-c", "--checks", help="Comma-separated list of checks (default: all)")
    ap.add_argument("--list", action="store_true", help="Show list of possible checks")
    ap.add_argument("-m", "--moduli", action="store_true",
                    help="Input file is list of RSA hex moduli")
    ap.add_argument("--crt-lines", action="store_true", help="Input file is list of base64 certs")
    ap.add_argument("--ssh-lines", action="store_true",
                    help="Input file is list of ssh public keys")
    ap.add_argument("--dkim", action="store_true", help="Scan DKIM records (in files)")
    ap.add_argument("--dkim-dns", action="store_true",
                    help="Scan DKIM DNS record (hostnames instead of files)")
    ap.add_argument("--dnssec", action="store_true", help="Scan DNSKEY/DNSSEC records (in files)")
    ap.add_argument("--jwk", action="store_true", help="Scan JSON Web Keys / Key Sets")
    ap.add_argument("-a", "--all", action="store_true", help="Show all keys")
    ap.add_argument("-w", "--warnings", action="store_true",
                    help="Enable extra warnings (keysize etc.)")
    ap.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    ap.add_argument("-j", "--json", action="store_true", help="JSON output")
    ap.add_argument("-q", "--quiet", action="store_true", help="Quiet output (for update commands)")
    ap.add_argument("-u", "--url", action="store_true", help="Show private key URL if possible")
    ap.add_argument("--update-bl", action="store_true", help="Update blocklist")
    ap.add_argument("-U", "--update-bl-and-urls", action="store_true",
                    help="Update blocklist and optional URL lookup list")
    ap.add_argument("--extrabl", help="comma-separated list of extra blocklist files")
    ap.add_argument("-t", "--tls", action="store_true",
                    help="Scan TLS (pass hostnames or IPs instead of files)")
    # default ports for https, smtps, imaps, pop3s, ldaps, ftps
    # and 8443 as a common non-default https port
    ap.add_argument("--tls-ports", default="443,465,636,990,993,995,8443",
                    help="TLS ports (comma-separated)")
    ap.add_argument("-s", "--ssh", action="store_true",
                    help="Scan SSH (pass hostnames or IPs instead of files)")
    ap.add_argument("--ssh-ports", default="22", help="SSH ports (comma-separated)")
    ap.add_argument("--version", action="version", version=__version__)
    args = ap.parse_args()

    if (args.moduli and args.crt_lines) or (args.moduli and args.ssh_lines) \
       or (args.ssh_lines and args.crt_lines):
        _errexit("Multiple input format parameters cannot be combined.")

    if (args.moduli or args.crt_lines or args.ssh_lines) and \
       (args.tls or args.ssh or args.dkim_dns):
        _errexit("Scan modes and input file modes cannot be combined.")

    if args.update_bl_and_urls:
        update_bl(lookup=True, quiet=args.quiet)
        sys.exit()
    if args.update_bl:
        update_bl(quiet=args.quiet)
        sys.exit()

    if args.list:
        for k, v in allchecks.items():
            print(f"{k}/{v['type']} keys: {v['desc']}")
        sys.exit()

    if not args.infiles:
        ap.print_help()
        sys.exit(1)

    if args.extrabl:
        extrabl = args.extrabl.split(",")
        for bl in extrabl:
            loadextrabl(bl)

    if args.checks:
        userchecks = args.checks.split(",")
        for c in userchecks:
            if c not in allchecks:
                _errexit(f"{c} is not a valid check")
    elif args.warnings:
        userchecks = allchecks.keys()
    else:
        userchecks = defaultchecks.keys()

    if args.tls:
        ports = [int(p) for p in args.tls_ports.split(",")]
        for host in args.infiles:
            for port in ports:
                keys = scantls(host, port, userchecks)
                for k in keys:
                    _printresults(k, f"tls:{host}:{port}", args)

    if args.ssh:
        ports = [int(p) for p in args.ssh_ports.split(",")]
        for host in args.infiles:
            for port in ports:
                keys = scanssh(host, port)
                for k in keys:
                    _printresults(k, f"ssh:{host}:{port}", args)

    if args.dkim_dns:
        try:
            import dns.resolver
        except ModuleNotFoundError:
            _errexit("DKIM DNS record scanning needs dnspython")
        for host in args.infiles:
            try:
                records = dns.resolver.resolve(host, "TXT").response
            except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
                _warnmsg(f"No TXT record found, {_esc(host)}")
                continue
            found = False
            for record in records.answer[-1]:
                dk = b"".join(record.strings).decode()
                key = parsedkim(dk)

                if not key:
                    continue
                found = True
                if isinstance(key, str):
                    r = {"type": "unparseable", "reason": key, "results": {}}
                else:
                    r = _checkkey(key, checks=userchecks)
                _printresults(r, host, args)
            if not found:
                _warnmsg(f"No DKIM/DomainKeys key in TXT record, {_esc(host)}")

    if args.jwk:
        for fn in args.infiles:
            with open(fn) as f:
                try:
                    j = json.load(f)
                except (json.decoder.JSONDecodeError, UnicodeDecodeError, ValueError):
                    _warnmsg(f"No valid JSON, {_esc(fn)}")
                    continue
            if isinstance(j, dict) and "kty" in j:
                r = checkjwk(j, checks=userchecks)
                _printresults(r, fn, args)
            elif isinstance(j, dict) and "keys" in j:
                for k in j["keys"]:
                    r = checkjwk(k, checks=userchecks)
                    _printresults(r, fn, args)
            else:
                _warnmsg(f"No JWK/JWKS, {_esc(fn)}")

    if args.ssh or args.tls or args.dkim_dns or args.jwk:
        sys.exit(_getret())

    if args.dnssec:
        dnskeyre = re.compile(r"[0-9]{1,3}\s+[0-9]{1,3}\s+[0-9]{1,3}\s+[A-Za-z0-9/+= ]+")

    for fn in args.infiles:
        if fn == "-":
            f = sys.stdin
        else:
            f = open(fn, errors="replace")

        if args.moduli or args.ssh_lines or args.crt_lines or args.dkim:
            lcount = 0

        if args.moduli:
            for line in f:
                try:
                    if line.startswith("Modulus="):
                        n = int(line[8:], 16)
                    else:
                        n = int(line, 16)
                except ValueError:
                    r = {"type": "unparseable", "results": {}}
                    _printresults(r, f"line {count}", args)
                    continue
                if n <= 0:
                    r = {"type": "unparseable", "results": {}}
                    _printresults(r, f"line {count}", args)
                    continue
                r = {"type": "rsa", "bits": n.bit_length()}
                r["results"] = checkrsa(n, checks=userchecks)
                _printresults(r, f"modulus {n:02x}", args)
                count += 1
                lcount += 1
        elif args.crt_lines:
            lno = 0
            for line in f:
                lno += 1
                desc = f"{fn}[{lno}]"
                ll = re.split("[,; ]", line.rstrip(), maxsplit=1)
                if len(ll) == 2:
                    desc += f" {ll[1]}"
                crt = PRECRT + ll[0] + POSTCRT
                r = checkcrt(crt, checks=userchecks)
                _printresults(r, desc, args)
                count += 1
                lcount += 1
        elif args.ssh_lines:
            lno = 0
            for sline in f:
                lno += 1
                line = sline.strip()
                if line == "" or line.startswith("#"):
                    continue
                desc = f"{fn}[{lno}]"
                if not line.startswith(("ssh-", "ecdsa-")):
                    if "ssh-" in line:
                        line = "ssh-" + line.split("ssh-", 1)[1]
                    elif "ecdsa-" in line:
                        line = "ecdsa-" + line.split("ecdsa-", 1)[1]
                    else:
                        _warnmsg(f"No SSH key in {desc}")
                        continue
                ll = line.split(" ", 2)
                if len(ll) == 3:
                    desc += f" {ll[2]}"
                r = checksshpubkey(line, checks=userchecks)
                _printresults(r, desc, args)
                count += 1
                lcount += 1
        elif args.dkim:
            lno = 0
            for line in f:
                lno += 1
                desc = f"{fn}[{lno}]"
                key = parsedkim(line)
                if not key:
                    continue
                if isinstance(key, str):
                    r = {"type": "unparseable", "reason": key, "results": {}}
                else:
                    r = _checkkey(key, checks=userchecks)
                _printresults(r, desc, args)
                count += 1
                lcount += 1
        elif args.dnssec:
            fcontent = f.read(MAXINPUTSIZE)

            keyrecs = dnskeyre.findall(fcontent)
            if not keyrecs:
                _warnmsg(f"No DNSSEC key found, {_esc(fn)}")
            for rec in keyrecs:
                r = checkdnskey(rec, checks=userchecks)
                _printresults(r, fn, args)
                count += 1
        else:
            fcontent = f.read(MAXINPUTSIZE)
            r = detectandcheck(fcontent, checks=userchecks)
            _printresults(r, fn, args)

        if args.moduli or args.ssh_lines or args.crt_lines or args.dkim:
            if lcount == 0:
                _printresults({"type": "notfound", "results": {}}, fn, args)

        if fn != "-":
            f.close()

    sys.exit(_getret())
